subroutine Bfield(Nz,Z,B,dB)

  implicit none

  integer Nz
  double precision Z(Nz), B(Nz), dB(Nz)

  integer ii
  double precision thick, Rmin, Rmax, zc, J, Bp(6,Nz), dBp(6,Nz)


! initialisation
  B=1.0d-3
  dB=0.0d0

  ii    = 1          ! coil number
  thick = 0.10d0     ! thickness in z
  Rmin  = 0.21d0     ! inner radius
  Rmax  = 0.61d0     ! outer radius
  zc    = 3.05d0     ! z-coordinate of coil centre
  J     = 768d4      ! current density in coil 
  call Bcoil(Nz,Z,Bp(ii,:),dBp(ii,:),zc,thick,Rmin,Rmax,J)

  ii    = 2          ! coil number
  thick = 0.10d0     ! thickness in z
  Rmin  = 0.60d0     ! inner radius
  Rmax  = 0.80d0     ! outer radius
  zc    = 2.35d0     ! z-coordinate of coil centre
  J     = 100d4      ! current density in coil 
  call Bcoil(Nz,Z,Bp(ii,:),dBp(ii,:),zc,thick,Rmin,Rmax,J)

  ii    = 3          ! coil number
  thick = 0.06d0     ! thickness in z
  Rmin  = 0.60d0     ! inner radius
  Rmax  = 0.80d0     ! outer radius
  zc    = 1.57d0     ! z-coordinate of coil centre
  J     = 100d4      ! current density in coil 
  call Bcoil(Nz,Z,Bp(ii,:),dBp(ii,:),zc,thick,Rmin,Rmax,J)

  ii    = 4          ! coil number
  thick = 0.06d0     ! thickness in z
  Rmin  = 0.60d0     ! inner radius
  Rmax  = 0.80d0     ! outer radius
  zc    = 0.81d0     ! z-coordinate of coil centre
  J     = 100d4      ! current density in coil 
  call Bcoil(Nz,Z,Bp(ii,:),dBp(ii,:),zc,thick,Rmin,Rmax,J)

  ii    = 5          ! coil number
  thick = 0.06d0     ! thickness in z
  Rmin  = 0.60d0     ! inner radius
  Rmax  = 0.80d0     ! outer radius
  zc    = 0.15d0     ! z-coordinate of coil centre
  J     = 100d4      ! current density in coil 
  call Bcoil(Nz,Z,Bp(ii,:),dBp(ii,:),zc,thick,Rmin,Rmax,J)

  ii    = 6          ! coil number
  thick = 0.06d0     ! thickness in z
  Rmin  = 0.60d0     ! inner radius
  Rmax  = 0.80d0     ! outer radius
  zc    =-0.61d0     ! z-coordinate of coil centre
  J     = 100d4      ! current density in coil 
  call Bcoil(Nz,Z,Bp(ii,:),dBp(ii,:),zc,thick,Rmin,Rmax,J)

  B  = sum(Bp,1)
  dB = sum(dBp,1)


end subroutine Bfield

!------------------------------------------------------------------------

subroutine Bcoil(Nz,Z,B,dB,zc,thick,Rmin,Rmax,J)

  implicit none

  integer Nz
  double precision Z(Nz), B(Nz), dB(Nz)

  double precision pi, thick, zc, Rmin, Rmax, b1(Nz), b2(Nz), J, &
       bp1(Nz), bp11(Nz), bp12(Nz), bp13(Nz), bp14(Nz), &
       bp2(Nz), bp21(Nz), bp22(Nz), bp23(Nz), bp24(Nz)

  parameter (pi=3.141592653589793d0)

! Compute B
  b1 = (Z+0.5d0*thick-zc) * &
       log( (Rmax + sqrt(Rmax**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0)) / &
       (Rmin + sqrt(Rmin**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0)) )

  b2 = (Z-0.5d0*thick-zc) * &
       log( (Rmax + sqrt(Rmax**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0)) / &
       (Rmin + sqrt(Rmin**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0)) )


  B = (2.0d0*pi*1.0d-7)*J*(b1 - b2)


! Compute dB/dz
  bp11 =  log( (Rmax + sqrt(Rmax**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0)) / &
       (Rmin + sqrt(Rmin**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0)) )

  bp12 = (Z+0.5d0*thick-zc) * &
       (Rmin + sqrt(Rmin**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0)) / &
       (Rmax + sqrt(Rmax**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0))

  bp13 = (Z+0.5d0*thick-zc) / &
       ((Rmin + sqrt(Rmin**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0)) * &
       sqrt(Rmax**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0) )

  bp14 = (Z+0.5d0*thick-zc) * &
       (Rmax + sqrt(Rmax**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0)) / &
       ((Rmin + sqrt(Rmin**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0))**2.0d0 * &
       sqrt(Rmin**2.0d0 + (Z+0.5d0*thick-zc)**2.0d0) )

  bp1 = bp11 + bp12*(bp13 - bp14)

  bp21 = log( (Rmax + sqrt(Rmax**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0)) / &
       (Rmin + sqrt(Rmin**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0)) )

  bp22 = (Z-0.5d0*thick-zc) * &
       (Rmin + sqrt(Rmin**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0)) / &
       (Rmax + sqrt(Rmax**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0))

  bp23 = (Z-0.5d0*thick-zc) / &
       ((Rmin + sqrt(Rmin**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0)) * &
       sqrt(Rmax**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0) )

  bp24 = (Z-0.5d0*thick-zc) * &
       (Rmax + sqrt(Rmax**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0)) / &
       ((Rmin + sqrt(Rmin**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0))**2.0d0 * &
       sqrt(Rmin**2.0d0 + (Z-0.5d0*thick-zc)**2.0d0) )

  bp2 = bp21 + bp22*(bp23 - bp24)


  dB = (2.0d0*pi*1.0d-7)*J*(bp1 - bp2)


end subroutine Bcoil

!------------------------------------------------------------------------

subroutine GetGravity(Nz,Z,gravity)

  implicit none

  integer Nz

  double precision Z(Nz), gravity(Nz)

  gravity = 0.0d0

end subroutine GetGravity

!------------------------------------------------------------------------
