%GENERAL PARAMETERS
dt = 7.0e-11;                 % timestep
Niter=200000;                % number of iterations to perform
dump_period_fields = 50;      % number of iterations between field dumps
fields_per_file = 200;        % number of field dumps to gather in each file
dump_period_distr = 500000;   % number of iterations between distribution dumps
dump_period_distr_1v = 100000;% number of iterations between f(z,vz) dumps
dump_start = 1;          % iteration at which dumping shall commence
shift_test_period = 10;  % number of time steps between testing for the
                         % need to shift
Nz=1400;                 % number of grid points in space
zmin=0.0;                % left hand boundary
zmax=3.0e0;              % right hand boundary
Nspecies=4;              % number of species
const_a = 162.66e0;      % epsilon_r=max((omegap*dt*const_a)**2,1.0d0)
BC_Poisson=1;            % Boundary condition for Poisson's equation.
                         % 1: V(right hand side) = voltage
                         % 2: E(left hand side) = E0
                         % 3: E(right hand side) = E0
voltage = 10.0e0;        % voltage across the system (potential of right
                         % hand side)
initialiser = 1;         % your choice of initialisation routine
                         % 1: original version
E0=0.0e0;                % Field at left hand boundary when BC_Poisson=2.
startfromdumpfile='yes'; % if 'yes' the distribution is read from the
                         % dumps directory
dump_period_dump=10000;  % number of iterations between dumps you can
                         % start from
exit_after_dump='no';   % whether execution shall stop after the first dump
transffilename='polesnm2level01d.dat'; % Name of transform file
voltagefilename='no file here';     % Name of voltage control file. If
                                    % missing or file non-existent a
                                    % constant voltage=voltage is used.
%END

%SPECIES 1: magnetospheric electrons
Nvz=100;               % number of grid points in velocity
vzmin =  -1.5e7;       % minimum velocity - vz0
vzmax =   1.5e7;       % maximum velocity - vz0
vzshifting = 'off';
Nmu = 50;              % number of grid points in magnetic moment
mumin = 0;             % minimum magnetic moment
mumax = 6.0e-16;       % maximum magnetic moment
muexp = 4; % exponent for nonuniform mu-grid mu~([1:Nmu]/Nmu)^muexp
mass = 9.10938215e-31; % if mass is NaN, an infinite mass neutralising
                       % background is assumed
charge = -1.602176487e-19;
relativistic = 'no';   % if 'yes', the particle mass is computed
                       % according to the theory of relativity.
n0 = 1.0e14;      % initial density maximum
vfromfile = 'no'; % If 'yes', load initial vz from file. The global vz0
                  % will still be the reference for velocity shifting.
vz0 = 0.0e0;      % drift velocity
Tfromfile = 'no'; % If 'yes', load T from file and ignore kTz and kTp.
kTz = 1.0e1;      % voltage equivalent to temperature kTz/e. f~exp(mv^2/(2kTz))
kTp = 1.0e1;      % voltage equivalent to temperature kTp/e. f~exp(mv^2/(2kTp))
n0L = 1.0e14;     % density at left hand boundary
vz0L = 0.0e6;     % drift velocity at left hand boundary
kTzL = 1.0e1;     % parallel temperature at left hand boundary
kTpL = 1.0e1;     % perpendicular temperature at left hand boundary
lossconeL = 'no'; % if 'yes' the left hand boundary loss cone is empty.
n0R =  0.0e9;     % density at right hand boundary
vz0R = 0.0e5;     % drift velocity at right hand boundary
kTzR = 0.2e0;     % parallel temperature at right hand boundary
kTpR = 0.2e0;     % perpendicular temperature at right hand boundary
lossconeR = 'no'; % if 'yes' the right hand boundary loss cone is empty.
%END

%SPECIES 2: magnetospheric neon
Nvz=400;               % number of grid points in velocity
vzmin =   -4.0e4;      % minimum velocity - vz0
vzmax =    4.0e4;      % maximum velocity - vz0
vzshifting = 'off';
Nmu = 50;              % number of grid points in magnetic moment
mumin = 0;             % minimum magnetic moment
mumax = 2.0e-18;       % maximum magnetic moment
muexp = 4; % exponent for nonuniform mu-grid mu~([1:Nmu]/Nmu)^muexp
mass = 3.350826e-26;   % if mass is NaN, an infinite mass neutralising
                       % background is assumed
%mass = NaN;
charge = 1.602176487e-19;
relativistic = 'no';   % if 'yes', the particle mass is computed
                       % according to the theory of relativity.
n0 = 1.0e14;      % initial density maximum
vz0 = 0.0e0;      % drift velocity
Tfromfile = 'no'; % if 'yes', load T from file and ignore kTz and kTp.
kTz = 3.0e-2;     % voltage equivalent to temperature kTz/e. f~exp(mv^2/(2kT))
kTp = 3.0e-2;     % voltage equivalent to temperature kTp/e. f~exp(mv^2/(2kT))
n0L = 1.0e14;     % density at left hand boundary
vz0L = 0.0;       % drift velocity at left hand boundary
kTzL = 3.0e-2;    % parallel temperature at left hand boundary
kTpL = 3.0e-2;    % perpendicular temperature at left hand boundary
lossconeL = 'no'; % if 'yes' the left hand boundary loss cone is empty.
n0R = 0.0e9;      % density at right hand boundary
vz0R = 0.0e0;     % drift velocity at right hand boundary
kTzR = 0.2e0;     % parallel temperature at right hand boundary
kTpR = 0.2e0;     % perpendicular temperature at right hand boundary
lossconeR = 'no'; % if 'yes' the right hand boundary loss cone is empty.
%END

%SPECIES 3: ionospheric electrons
Nvz=100;               % number of grid points in velocity
vzmin =  -3.0e6;       % minimum velocity - vz0
vzmax =   3.0e6;       % maximum velocity - vz0
vzshifting = 'off';
Nmu = 20;              % number of grid points in magnetic moment
mumin = 0;             % minimum magnetic moment
mumax = 1.0e-18;       % maximum magnetic moment
muexp = 4;             % exponent for nonuniform mu-grid mu~([1:Nmu]/Nmu)^muexp
mass = 9.10938215e-31; % if mass is NaN, an infinite mass neutralising
                       % background is assumed
charge = -1.602176487e-19;
relativistic = 'no';   % if 'yes', the particle mass is computed
                       % according to the theory of relativity.
n0 = 1.0e15;           % initial density maximum
vfromfile = 'no';      % If 'yes', load initial vz from file. The global vz0
                       % will still be the reference for velocity shifting.
vz0 = 0.0e0;           % drift velocity
Tfromfile = 'no';      % If 'yes', load T from file and ignore kTz and kTp.
kTz = 0.2e0;      % voltage equivalent to temperature kTz/e. f~exp(mv^2/(2kTz))
kTp = 0.2e0;      % voltage equivalent to temperature kTp/e. f~exp(mv^2/(2kTp))
n0L = 0.0e5;      % density at left hand boundary
vz0L = 0.0e6;     % drift velocity at left hand boundary
kTzL = 1.0e1;     % parallel temperature at left hand boundary
kTpL = 1.0e1;     % perpendicular temperature at left hand boundary
lossconeL = 'no'; % if 'yes' the left hand boundary loss cone is empty.
n0R = 1.0e15;     % density at right hand boundary
vz0R = 0.0e5;     % drift velocity at right hand boundary
kTzR = 0.2e0;     % parallel temperature at right hand boundary
kTpR = 0.2e0;     % perpendicular temperature at right hand boundary
lossconeR = 'no'; % if 'yes' the right hand boundary loss cone is empty.
%END

%SPECIES 4: ionospheric sodium
Nvz=100;               % number of grid points in velocity
vzmin =   -2.0e4;      % minimum velocity - vz0
vzmax =    2.0e4;      % maximum velocity - vz0
vzshifting = 'on';
Nmu = 20;              % number of grid points in magnetic moment
mumin = 0;             % minimum magnetic moment
mumax = 1.0e-18;       % maximum magnetic moment
muexp = 4;             % exponent for nonuniform mu-grid mu~([1:Nmu]/Nmu)^muexp
mass = 3.817449e-26;   % if mass is NaN, an infinite mass neutralising
                       % background is assumed
%mass = NaN;
charge = 1.602176487e-19;
relativistic = 'no';   % if 'yes', the particle mass is computed
                       % according to the theory of relativity.
n0 =  1.0e15;     % initial density maximum
vz0 = 0.0e0;      % drift velocity
Tfromfile = 'no'; % if 'yes', load T from file and ignore kTz and kTp.
kTz = 0.2e0;      % voltage equivalent to temperature kTz/e. f~exp(mv^2/(2kT))
kTp = 0.2e0;      % voltage equivalent to temperature kTp/e. f~exp(mv^2/(2kT))
n0L = 0.0e5;      % density at left hand boundary
vz0L = 0.0;       % drift velocity at left hand boundary
kTzL = 1.0e1;     % parallel temperature at left hand boundary
kTpL = 1.0e1;     % perpendicular temperature at left hand boundary
lossconeL = 'no'; % if 'yes' the left hand boundary loss cone is empty.
n0R = 1.0e15;     % density at right hand boundary
vz0R = 0.0e0;     % drift velocity at right hand boundary
kTzR = 0.2e0;     % parallel temperature at right hand boundary
kTpR = 0.2e0;     % perpendicular temperature at right hand boundary
lossconeR = 'no'; % if 'yes' the right hand boundary loss cone is empty.
%END
